import type { BaseQueryFn, FetchBaseQueryError } from '@reduxjs/toolkit/query';
import { Mutex } from 'async-mutex';

import { getAuthAttribute, setAuthAttribute, setAuthTokens } from '@procyon/auth/slice';
import { AUTH_REDUCER_NAME, AuthRootState } from '@procyon/auth/types';
import { isJSON } from '@procyon/utils';

import { buildUrl } from './utils';

export type BaseQueryParams = {
	authentication?: { headerName: string; tokenType: string };
	baseUrl: string;
};

export type BaseQueryFetchParams<Body> = {
	body?: Body;
	headers?: Record<string, string | number | boolean>;
	method?: string;
	params?: any;
	url: string;
};

export type BaseQueryWithRefreshParams = {
	authentication: {
		baseUrl: string;
		clientId: string;
		headerName: string;
		redirectLoginUri: string;
		tokenEndpoint: string;
		tokenType: string;
	};
	baseUrl: string;
};

const mutex = new Mutex();

export const baseQuery =
	<Response = { data: unknown }, RequestBody = any>({
		baseUrl,
		authentication,
	}: BaseQueryParams): BaseQueryFn<BaseQueryFetchParams<RequestBody>, Response, FetchBaseQueryError> =>
	async ({ url, method = 'GET', params, body, headers }, { getState }) => {
		// Authentication
		if (authentication && (getState() as AuthRootState).procyon[AUTH_REDUCER_NAME] !== undefined) {
			const isAuthenticated = getAuthAttribute(getState() as AuthRootState, 'authenticated');
			if (isAuthenticated) {
				const tokenValue = getAuthAttribute(getState() as AuthRootState, 'accessToken');
				headers = {
					...headers,
					[authentication.headerName]: `${authentication.tokenType} ${tokenValue?.toString()}`.trim(),
				};
			}
		}
		// Call endpoint
		try {
			const response = await fetch(`${baseUrl}${buildUrl(url, params)}`, {
				headers: {
					'Content-Type': 'application/json',
					...headers,
				},
				method: method.toUpperCase(),
				...(body ? { body: new URLSearchParams(body) } : {}),
			});
			// Resolve response
			const json = await response.text().then((text: string) => (isJSON(text) ? JSON.parse(text) : text));
			if (!response.ok) throw response;
			return { data: json };
		} catch (err) {
			return { error: { status: err.status, error: err } };
		}
	};

export const baseQueryWithRefresh =
	<Response = { data: unknown }, RequestBody = any>({
		baseUrl,
		authentication,
	}: BaseQueryWithRefreshParams): BaseQueryFn<BaseQueryFetchParams<RequestBody>, Response, FetchBaseQueryError> =>
	async (args, api, extraOptions) => {
		const authReducer: any = (api.getState() as AuthRootState).procyon[AUTH_REDUCER_NAME];
		const accessTokenExpired = new Date(authReducer.accessExpiresTime) < new Date();
		if (accessTokenExpired) {
			await mutex.waitForUnlock();
			const refreshTokenExpired = new Date(authReducer.refreshExpiresTime) < new Date();
			if (!refreshTokenExpired) {
				if (!mutex.isLocked()) {
					const release = await mutex.acquire();
					try {
						const { data: tokens } = await (baseQuery({ baseUrl: authentication.baseUrl, authentication }) as any)(
							{
								url: authentication.tokenEndpoint,
								method: 'post',
								headers: { 'Content-Type': 'application/x-www-form-urlencoded' },
								body: {
									client_id: authentication.clientId,
									grant_type: 'refresh_token',
									refresh_token: authReducer.refreshToken,
								},
							},
							api,
							extraOptions
						);
						api.dispatch(
							setAuthTokens({
								accessToken: tokens.access_token,
								refreshToken: tokens.refresh_token,
								idToken: tokens.id_token,
								accessExpiresIn: tokens.expires_in,
								refreshExpiresIn: tokens.refresh_expires_in,
								sessionState: tokens.session_state,
							})
						);
					} finally {
						release();
					}
				}
			} else {
				api.dispatch(setAuthAttribute({ authenticated: false }));
				window.location.assign(authentication.redirectLoginUri);
				return null as any;
			}
		}
		return await baseQuery<Response, RequestBody>({ baseUrl, authentication })(args, api, extraOptions);
	};
